/* -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/* Procman - dialogs
 * Copyright (C) 2001 Kevin Vandersloot
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 */


#include <config.h>

#include <glib/gi18n.h>

#include <signal.h>
#include <string.h>
#include <stdio.h>

#include "servicesdialogs.h"
#include "servicestable.h"
#include "prettytable.h"
#include "servicesactions.h"
#include "util.h"
#include "gsm_gnomesu.h"
#include "gsm_gksu.h"
#include "gsm_pkexec.h"
#include "cgroups.h"



static
void dialog_button_pressed (GtkDialog *dialog, gint id, gpointer data)
{
    struct ServicesActionArgs *kargs = static_cast<ServicesActionArgs*>(data);
    gtk_widget_destroy (GTK_WIDGET (dialog));
    if (id == GTK_RESPONSE_OK)
    do_process (kargs->app, kargs->arg_value);
    g_free (kargs);
}

void servicesdialog_create_dialog(GsmApplication *app, int signal) {

    GtkMessageDialog *dialog;
    GtkWidget *confirm_button;

    gchar *primary, *secondary, *button_text;
    struct ServicesActionArgs *kargs;

    kargs = g_new(ServicesActionArgs, 1);
    kargs->app = app;
    kargs->arg_value = signal;
    gint selected_count = gtk_tree_selection_count_selected_rows (app->services_selection);


    if ( selected_count == 1 ) {
        ServicesInfo *selected_service = NULL;
        // get the last selected row
        gtk_tree_selection_selected_foreach (app->services_selection, get_last_selected_services,
                                             &selected_service);

        std::string *service_name = &selected_service->name;

        switch (signal) {
            case SERVICESMAN_ACTION_START:
                /*xgettext: primary alert message for start single process*/
                primary = g_strdup_printf (_("Are you sure you want to start the selected service “%s”?"),
                                           service_name->c_str());
                break;
            case SERVICESMAN_ACTION_STOP:
                /*xgettext: primary alert message for stop single process*/
                primary = g_strdup_printf (_("Are you sure you want to stop the selected service “%s”?"),
                                           service_name->c_str());
                break;
            case SERVICESMAN_ACTION_RESTART:
                /*xgettext: primary alert message for stop single process*/
                primary = g_strdup_printf (_("Are you sure you want to resstart the selected service “%s”?"),
                                           service_name->c_str());
                break;
            case SERVICESMAN_ACTION_ENABLE:
                /*xgettext: primary alert message for enable single process*/
                primary = g_strdup_printf (_("Are you sure you want to enable the selected service “%s” ?"),
                                           service_name->c_str());
                break;
            case SERVICESMAN_ACTION_DISABLE:
                /*xgettext: primary alert message for disable single processes*/
                primary = g_strdup_printf (_("Are you sure you want to disable the selected service “%s” ?"),
                                           service_name->c_str());
                break;
        }
    } else {
        switch (signal) {
            case SERVICESMAN_ACTION_START:
                /*xgettext: primary alert message for start multiple processes*/
                primary = g_strdup_printf (ngettext("Are you sure you want to start the selected service?",
                                                    "Are you sure you want to start the %d selected service?", selected_count),
                                           selected_count);
                break;
            case SERVICESMAN_ACTION_STOP:
                /*xgettext: primary alert message for stop multiple processes*/
                primary = g_strdup_printf (ngettext("Are you sure you want to stop the selected service?",
                                                    "Are you sure you want to stop the %d selected service?", selected_count),
                                           selected_count);
                break;
            case SERVICESMAN_ACTION_RESTART:
                /*xgettext: primary alert message for stop multiple processes*/
                primary = g_strdup_printf (ngettext("Are you sure you want to restart the selected service?",
                                                    "Are you sure you want to restart the %d selected service?", selected_count),
                                           selected_count);
                break;
            case SERVICESMAN_ACTION_ENABLE:
                /*xgettext: primary alert message for enable multiple processes*/
                primary = g_strdup_printf (ngettext("Are you sure you want to enable the selected service?",
                                                    "Are you sure you want to enable the %d selected service?", selected_count),
                                           selected_count);
                break;
            case SERVICESMAN_ACTION_DISABLE:
                /*xgettext: primary alert message for disable multiple processes*/
                primary = g_strdup_printf (ngettext("Are you sure you want to disable the selected service?",
                                                    "Are you sure you want to disable the %d selected service?", selected_count),
                                           selected_count);
                break;
        }
    }

    switch (signal) {
        case SERVICESMAN_ACTION_START:
            /*xgettext: secondary alert message*/
            secondary = _("Start a service unknown may destroy data, break the "
                          "session or introduce a security risk. "
                          "Only do this when you know what are you doing.");
            button_text = ngettext("_Start Service", "_Start Service", selected_count);
            break;
        case SERVICESMAN_ACTION_STOP:
            /*xgettext: secondary alert message*/
            secondary = _("Stop a service may destroy data, break the "
                          "session or introduce a security risk. "
                          "Only do this when you know what are you doing.");
            button_text = ngettext("_Stop Service", "_Stop Service", selected_count);
            break;
        case SERVICESMAN_ACTION_RESTART:
            /*xgettext: secondary alert message*/
            secondary = _("Restart a service may destroy data, break the "
                          "session or introduce a security risk. "
                          "Only do this when you know what are you doing.");
            button_text = ngettext("_Restart Service", "_Restart Service", selected_count);
            break;
        case SERVICESMAN_ACTION_ENABLE:
            /*xgettext: secondary alert message*/
            secondary = _("Enabled service will auto load when system loaded."
                          "Unknown service may do harm to your computer.");
            button_text = ngettext("_Enable Service", "_Enable Service", selected_count);
            break;
        case SERVICESMAN_ACTION_DISABLE:
            /*xgettext: secondary alert message*/
            secondary = _("Disabled service will not auto load."
                          "Please don't disable important service.");
            button_text = ngettext("_Disable Service", "_Disable Service", selected_count);
            break;
    }

    switch (signal) {
        case SERVICESMAN_ACTION_START:
        case SERVICESMAN_ACTION_ENABLE:
        case SERVICESMAN_ACTION_RESTART:
            dialog = GTK_MESSAGE_DIALOG (gtk_message_dialog_new (GTK_WINDOW (app->main_window),
                                                                 static_cast<GtkDialogFlags>(GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT),
                                                                 GTK_MESSAGE_INFO,
                                                                 GTK_BUTTONS_NONE,
                                                                 "%s",
                                                                 primary));
            break;
        case SERVICESMAN_ACTION_STOP:
        case SERVICESMAN_ACTION_DISABLE:
        default:
            dialog = GTK_MESSAGE_DIALOG (gtk_message_dialog_new (GTK_WINDOW (app->main_window),
                                                                 static_cast<GtkDialogFlags>(GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT),
                                                                 GTK_MESSAGE_WARNING,
                                                                 GTK_BUTTONS_NONE,
                                                                 "%s",
                                                                 primary));
    }

    g_free (primary);

    gtk_message_dialog_format_secondary_text (dialog,
                                              "%s",
                                              secondary);

    gtk_dialog_add_button (GTK_DIALOG (dialog),
                           _("_Cancel"), GTK_RESPONSE_CANCEL);

    confirm_button = gtk_dialog_add_button (GTK_DIALOG (dialog),
                                            button_text, GTK_RESPONSE_OK);
    gtk_style_context_add_class (gtk_widget_get_style_context (confirm_button),
                                 GTK_STYLE_CLASS_DESTRUCTIVE_ACTION);

    gtk_dialog_set_default_response (GTK_DIALOG (dialog),
                                     GTK_RESPONSE_CANCEL);

    g_signal_connect (G_OBJECT (dialog), "response",
                      G_CALLBACK (dialog_button_pressed), kargs);

    gtk_widget_show_all (GTK_WIDGET (dialog));
}


static char *
servicesman_action_to_command(ServicesmanActionType type,
                              const std::string& service_name) {
    switch (type) {
        case SERVICESMAN_ACTION_START:
            return g_strdup_printf("Start %s", service_name.c_str());
        case SERVICESMAN_ACTION_STOP:
            return g_strdup_printf("Stop %s", service_name.c_str());
        case SERVICESMAN_ACTION_ENABLE:
            return g_strdup_printf("Enable %s", service_name.c_str());
        case SERVICESMAN_ACTION_DISABLE:
            return g_strdup_printf("Disable %s", service_name.c_str());
        default:
            g_assert_not_reached();
    }
}


/*
 * type determines whether if dialog is for killing services or renice.
 * type == PROCMAN_ACTION_KILL,   extra_value -> signal to send
 * type == PROCMAN_ACTION_RENICE, extra_value -> new priority.
 */
gboolean
servicesdialog_create_root_password_dialog(ServicesmanActionType type,
                                           GsmApplication *app,
                                           std::string name,
                                           gint extra_value) {
    char *command;
    gboolean ret = FALSE;

    command = servicesman_action_to_command(type, name);

    if (procman_has_pkexec())
        ret = gsm_pkexec_create_root_password_dialog(command);
    else if (procman_has_gksu())
        ret = gsm_gksu_create_root_password_dialog(command);
    else if (procman_has_gnomesu())
        ret = gsm_gnomesu_create_root_password_dialog(command);

    g_free(command);
    return ret;
}
