/* -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/* services properties dialog
 * Copyright (C) 2010 Krishnan Parthasarathi <krishnan.parthasarathi@gmail.com>
 *                    Robert Ancell <robert.ancell@canonical.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <config.h>

#include <glib/gi18n.h>
#include <glibtop/procmem.h>
#include <glibtop/procmap.h>
#include <glibtop/procstate.h>

#include "application.h"
#include "servicesproperties.h"
#include "servicestable.h"
#include "util.h"
#include "legacy/e_date.h"

enum
{
    SERVICES_COL_PROP = 0,
    SERVICES_COL_VAL,
    NUM_COLS,
};

typedef struct _services_arg {
    const gchar *prop;
    gchar *val;
} services_arg;

static gchar*
format_memsize(guint64 size)
{
    if (size == 0)
        return g_strdup(_("N/A"));
    else
        return g_format_size_full(size, G_FORMAT_SIZE_IEC_UNITS);
}

static void
fill_services_properties (GtkTreeView *tree, ServicesInfo *info)
{
    guint i;
    GtkListStore *store;
    
    if (!info)
        return;

    services_arg services_props[] = {
        { N_("Services Name"), g_strdup_printf("%s", info->name.c_str())},
        { N_("Load"), g_strdup_printf("%s", info->load.c_str())},
        { N_("Activate"), g_strdup_printf("%s", info->activate.c_str())},
        { N_("Sub"), g_strdup_printf("%s", info->sub.c_str())},
        { N_("Description"), g_strdup_printf("%s", info->description.c_str())},
        { NULL, NULL}
    };

    store = GTK_LIST_STORE(gtk_tree_view_get_model(tree));
    for (i = 0; services_props[i].prop; i++) {
        GtkTreeIter iter;

        if (!gtk_tree_model_iter_nth_child (GTK_TREE_MODEL(store), &iter, NULL, i)) {
            gtk_list_store_append(store, &iter);
            gtk_list_store_set(store, &iter, SERVICES_COL_PROP, gettext(services_props[i].prop), -1);
        }

        gtk_list_store_set(store, &iter, SERVICES_COL_VAL, g_strstrip(services_props[i].val), -1);
        g_free(services_props[i].val);
    }
}

static GtkTreeView *
create_servicesproperties_tree (GsmApplication *app, ServicesInfo *info)
{
    GtkTreeView *tree;
    GtkListStore *model;
    GtkTreeViewColumn *column;
    GtkCellRenderer *cell;
    gint i;

    model = gtk_list_store_new (NUM_COLS,
                                G_TYPE_STRING,	/* Property */
                                G_TYPE_STRING	/* Value */
        );

    tree = GTK_TREE_VIEW (gtk_tree_view_new_with_model (GTK_TREE_MODEL (model)));
    g_object_unref (G_OBJECT (model));

    for (i = 0; i < NUM_COLS; i++) {
        cell = gtk_cell_renderer_text_new ();

        column = gtk_tree_view_column_new_with_attributes (NULL,
                                                           cell,
                                                           "text", i,
                                                           NULL);
        gtk_tree_view_column_set_resizable (column, TRUE);
        gtk_tree_view_append_column (tree, column);
    }

    gtk_tree_view_set_headers_visible (tree, FALSE);
    fill_services_properties(tree, info);

    return tree;
}

static gboolean
servicesprop_timer (gpointer data)
{
    GtkTreeView *tree = static_cast<GtkTreeView*>(data);
    GtkTreeModel *model;

    model = gtk_tree_view_get_model (tree);
    g_assert(model);
    
    return TRUE;
}

static void
create_single_servicesproperties_dialog (GtkTreeModel *model, GtkTreePath *path,
                                     GtkTreeIter *iter, gpointer data)
{
    GsmApplication *app = static_cast<GsmApplication *>(data);

    GtkDialog *servicespropdialog;
    GtkBox *dialog_vbox, *vbox;
    GtkBox *cmd_hbox;
    gchar *label;
    GtkScrolledWindow *scrolled;
    GtkTreeView *tree;
    ServicesInfo *info;
    guint timer;

    gtk_tree_model_get (model, iter, SERVICES_COL_POINTER, &info, -1);

    if (!info)
        return;

    servicespropdialog = GTK_DIALOG (g_object_new (GTK_TYPE_DIALOG, 
                                               "use-header-bar", TRUE, NULL));

    label = g_strdup_printf( _("%s"), info->name.c_str());
    gtk_window_set_title (GTK_WINDOW (servicespropdialog), label);
    g_free (label);

    gtk_window_set_destroy_with_parent (GTK_WINDOW (servicespropdialog), TRUE);

    gtk_window_set_resizable (GTK_WINDOW (servicespropdialog), TRUE);
    gtk_window_set_default_size (GTK_WINDOW (servicespropdialog), 575, 400);
    gtk_container_set_border_width (GTK_CONTAINER (servicespropdialog), 5);

    vbox = GTK_BOX (gtk_dialog_get_content_area (GTK_DIALOG (servicespropdialog)));
    gtk_box_set_spacing (vbox, 2);
    gtk_container_set_border_width (GTK_CONTAINER (vbox), 5);

    dialog_vbox = GTK_BOX (gtk_box_new (GTK_ORIENTATION_VERTICAL, 6));
    gtk_container_set_border_width (GTK_CONTAINER (dialog_vbox), 5);
    gtk_box_pack_start (vbox, GTK_WIDGET (dialog_vbox), TRUE, TRUE, 0);

    cmd_hbox = GTK_BOX (gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 12));
    gtk_box_pack_start (dialog_vbox, GTK_WIDGET (cmd_hbox), FALSE, FALSE, 0);

    scrolled = GTK_SCROLLED_WINDOW (gtk_scrolled_window_new (NULL, NULL));
    gtk_scrolled_window_set_policy (scrolled,
                                    GTK_POLICY_AUTOMATIC,
                                    GTK_POLICY_AUTOMATIC);
    gtk_scrolled_window_set_shadow_type (scrolled,
                                         GTK_SHADOW_IN);

    tree = create_servicesproperties_tree (app, info);
    gtk_container_add (GTK_CONTAINER (scrolled), GTK_WIDGET (tree));

    gtk_box_pack_start (dialog_vbox, GTK_WIDGET (scrolled), TRUE, TRUE, 0);
    gtk_widget_show_all (GTK_WIDGET (scrolled));

    gtk_window_set_transient_for (GTK_WINDOW (servicespropdialog), GTK_WINDOW (GsmApplication::get()->main_window));
    gtk_widget_show_all (GTK_WIDGET (servicespropdialog));

    timer = g_timeout_add_seconds (5, servicesprop_timer, tree);
    g_object_set_data (G_OBJECT (tree), "timer", GUINT_TO_POINTER (timer));

}

void
create_servicesproperties_dialog (GsmApplication *app)
{
    gtk_tree_selection_selected_foreach (app->services_selection, create_single_servicesproperties_dialog,
                                         app);
}
