/* -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/* Procman tree view and services updating
 * Copyright (C) 2001 Kevin Vandersloot
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 */


#include <config.h>


#include <string.h>
#include <math.h>
#include <glib/gi18n.h>
#include <glib/gprintf.h>
#include <glibtop.h>
#include <glibtop/proclist.h>
#include <glibtop/procstate.h>
#include <glibtop/procio.h>
#include <glibtop/procmem.h>
#include <glibtop/procmap.h>
#include <glibtop/proctime.h>
#include <glibtop/procuid.h>
#include <glibtop/procargs.h>
#include <glibtop/prockernel.h>
#include <glibtop/mem.h>
#include <glibtop/swap.h>
#include <sys/stat.h>
#include <pwd.h>
#include <time.h>

#include <set>
#include <list>

#ifdef HAVE_WNCK
#define WNCK_I_KNOW_THIS_IS_UNSTABLE
#include <libwnck/libwnck.h>
#endif

#include "application.h"
#include "servicestable.h"
#include "prettytable.h"
#include "util.h"
#include "interface.h"
#include "selinux.h"
#include "settings-keys.h"
#include "cgroups.h"
#include "legacy/treeview.h"
#include "systemd.h"
#include "servicesutil.h"

#ifdef GDK_WINDOWING_X11
#include <gdk/gdkx.h>
#include <systemd/sd-bus.h>

#endif

ServicesInfo* ServicesList::find(const std::string name)
{
    auto iter = data.begin();
    for ( ; iter != data.end(); iter ++) {
        if ( strcmp(iter->first.c_str(), name.c_str()) == 0 &&
            iter->first.length() == name.length()) {
            return &iter->second;
        }
    }
    return NULL;
}

static void
cb_save_servicestree_state(gpointer, gpointer data)
{
    GsmApplication * const app = static_cast<GsmApplication *>(data);

    gsm_tree_view_save_state (app->services_tree);
}

static void
cb_servicestree_destroying (GtkTreeView *self, gpointer data)
{
    g_signal_handlers_disconnect_by_func (self,
                                          (gpointer) cb_save_servicestree_state,
                                          data);

    g_signal_handlers_disconnect_by_func (gtk_tree_view_get_model (self),
                                          (gpointer) cb_save_servicestree_state,
                                          data);
}

static gboolean
cb_services_tree_button_pressed (GtkWidget *widget, GdkEventButton *event, gpointer data)
{
    GsmApplication *app = (GsmApplication *) data;
    GtkTreePath *path;

    if (!gdk_event_triggers_context_menu ((GdkEvent *) event))
        return FALSE;

    if (!gtk_tree_view_get_path_at_pos (GTK_TREE_VIEW (app->services_tree), event->x, event->y, &path, NULL, NULL, NULL))
        return FALSE;

    if (!gtk_tree_selection_path_is_selected (app->services_selection, path)) {
        if (!(event->state & (GDK_CONTROL_MASK | GDK_SHIFT_MASK)))
            gtk_tree_selection_unselect_all (app->services_selection);
        gtk_tree_selection_select_path (app->services_selection, path);
    }

    gtk_tree_path_free (path);

    gtk_menu_popup_at_pointer (GTK_MENU (app->services_popup_menu), NULL);
    return TRUE;
}

static gboolean
cb_services_tree_popup_menu (GtkWidget *widget, gpointer data)
{
    GsmApplication *app = (GsmApplication *) data;

    gtk_menu_popup_at_pointer (GTK_MENU (app->services_popup_menu), NULL);

    return TRUE;
}

void
get_last_selected_services (GtkTreeModel *model, GtkTreePath *path,
                   GtkTreeIter *iter, gpointer data)
{
    ServicesInfo **info = (ServicesInfo**) data;

    gtk_tree_model_get (model, iter, SERVICES_COL_POINTER, info, -1);
}

static void
cb_services_row_selected (GtkTreeSelection *selection, gpointer data)
{
    GsmApplication *app = (GsmApplication *) data;

    app->services_selection = selection;

    ServicesInfo *selected_services = NULL;

    /* get the most recent selected services and determine if there are
    ** no selected services
    */
    gtk_tree_selection_selected_foreach (app->services_selection, get_last_selected_services,
                                         &selected_services);
    if (selected_services) {
        GVariant *priority;
        gint nice = selected_services->nice;
        if (nice < -7)
            priority = g_variant_new_int32 (-20);
        else if (nice < -2)
            priority = g_variant_new_int32 (-5);
        else if (nice < 3)
            priority = g_variant_new_int32 (0);
        else if (nice < 7)
            priority = g_variant_new_int32 (5);
        else
            priority = g_variant_new_int32 (19);

//        GAction *action = g_action_map_lookup_action (G_ACTION_MAP (app->main_window),
//                                                      "priority");
//
//        g_action_change_state (action, priority);
    }
    update_sensitivity(app);
}

static gint
cb_services_timeout (gpointer data)
{
    GsmApplication *app = (GsmApplication *) data;
    guint new_interval;

    servicestable_update (app);

    if (app->smooth_refresh->get(new_interval)) {
        app->timeout = g_timeout_add(new_interval,
                                     cb_services_timeout,
                                     app);
        return G_SOURCE_REMOVE;
    }

    return G_SOURCE_CONTINUE;
}

//static void
//cb_refresh_services_icons (GtkIconTheme *theme, gpointer data)
//{
//    GsmApplication *app = (GsmApplication *) data;
//
//    for (auto& v : app->services) {
//        app->services_pretty_table->set_icon(v.second);
//    }
//
//    servicestable_update (app);
//}

static gboolean
iter_matches_search_key (GtkTreeModel *model, GtkTreeIter *iter, const gchar *search_text)
{
    char *name;
    char *load;
    char *activate;
    char *sub;
    gboolean found;
    char *search_pattern;
    char **keys;
    Glib::RefPtr<Glib::Regex> regex;

    gtk_tree_model_get (model, iter,
                        SERVICES_COL_NAME, &name,
                        SERVICES_COL_LOAD, &load,
                        SERVICES_COL_ACTIVATE, &activate,
                        SERVICES_COL_SUB, &sub,
                        -1);


    keys = g_strsplit_set(search_text, " |", -1);
    search_pattern = g_strjoinv ("|", keys);
    try {
        regex = Glib::Regex::create(search_pattern, Glib::REGEX_CASELESS);
    } catch (const Glib::Error& ex) {
        regex = Glib::Regex::create(Glib::Regex::escape_string(search_pattern), Glib::REGEX_CASELESS);
    }

    found = (name && regex->match(name)) || (load && regex->match(load))
            || (activate && regex->match(activate)) || (sub && regex->match(sub));

    g_strfreev (keys);
    g_free (search_pattern);
    g_free (name);
    g_free (load);
    g_free (activate);
    g_free (sub);

    return found;
}

static gboolean
services_visibility_func (GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
    GsmApplication * const app = static_cast<GsmApplication *>(data);
    const gchar * search_text = app->services_search_entry == NULL ? "" : gtk_entry_get_text (GTK_ENTRY (app->services_search_entry));
    GtkTreePath *tree_path = gtk_tree_model_get_path (model, iter);

    if (strcmp (search_text, "") == 0) {
        gtk_tree_path_free (tree_path);
        return TRUE;
    }

	// in case we are in dependencies view, we show (and expand) rows not matching the text, but having a matching child
    gboolean match = false;
    if (app->settings->get_boolean (GSM_SETTING_SHOW_DEPENDENCIES)) {
        GtkTreeIter child;
        if (gtk_tree_model_iter_children (model, &child, iter)) {
            gboolean child_match = FALSE;
            do {
                child_match = services_visibility_func (model, &child, data);
            } while (gtk_tree_model_iter_next (model, &child) && !child_match);
            match = child_match;
        }

        match |= iter_matches_search_key (model, iter, search_text);
        // TODO auto-expand items not matching the search string but having matching children
        // complicated because of treestore nested in treemodelfilter nested in treemodelsort
        // expand to path requires the path string in the treemodelsort, but tree_path is the path in the double nested treestore
        //if (match && (strlen (search_text) > 0)) {
        //    gtk_tree_view_expand_to_path (GTK_TREE_VIEW (app->services_tree), tree_path);
        //}

    } else {
        match = iter_matches_search_key (model, iter, search_text);
    }

    gtk_tree_path_free (tree_path);

    return match;
}

static void
servicestable_clear_tree (GsmApplication * const app)
{
    GtkTreeModel *model;

    model =  gtk_tree_model_filter_get_model (GTK_TREE_MODEL_FILTER (
             gtk_tree_model_sort_get_model(GTK_TREE_MODEL_SORT (
             gtk_tree_view_get_model (GTK_TREE_VIEW(app->services_tree))))));

    gtk_tree_store_clear (GTK_TREE_STORE (model));

    servicestable_free_table (app);

    update_sensitivity(app);
}

static void
cb_show_services_dependencies_changed(Gio::Settings& settings, Glib::ustring key, GsmApplication* app) {
    if (app->timeout) {
        gtk_tree_view_set_show_expanders (GTK_TREE_VIEW (app->services_tree),
                                          settings.get_boolean (GSM_SETTING_SHOW_DEPENDENCIES));

        servicestable_clear_tree (app);
        servicestable_update (app);
    }
}

static void
cb_show_whose_services_changed(Gio::Settings& settings, Glib::ustring key, GsmApplication* app) {
    if (app->timeout) {
        servicestable_clear_tree (app);
        servicestable_update (app);
    }
}

GsmTreeView *
servicestable_new (GsmApplication * const app)
{
    GsmTreeView *servicestree;
    GtkTreeStore *model;
    GtkTreeModelFilter *model_filter;
    GtkTreeModelSort *model_sort;

    GtkTreeViewColumn *column;
    GtkCellRenderer *cell_renderer;
    const gchar *titles[] = {
        N_("Services Name"),
        N_("Load"),
        N_("Activate"),
        N_("Sub"),
        N_("Description"),
        "POINTER"
    };

    gint i;
    auto settings = g_settings_get_child (app->settings->gobj (), GSM_SETTINGS_CHILD_SERVICES);
    model = gtk_tree_store_new (SERVICES_NUM_COLUMNS,
                                G_TYPE_STRING,      /* Services Name */
                                G_TYPE_STRING,      /* Load         */
                                G_TYPE_STRING,        /* Activate       */
                                G_TYPE_STRING,       /* Sub      */
                                G_TYPE_STRING,        /* Information           */
                                G_TYPE_POINTER,     /* ServicesInfo     */
                                G_TYPE_STRING       /* Sexy tooltip */
        );

    model_filter = GTK_TREE_MODEL_FILTER (gtk_tree_model_filter_new (GTK_TREE_MODEL (model), NULL));

    gtk_tree_model_filter_set_visible_func(model_filter, services_visibility_func, app, NULL);

    model_sort = GTK_TREE_MODEL_SORT (gtk_tree_model_sort_new_with_model (GTK_TREE_MODEL (model_filter)));

    servicestree = gsm_tree_view_new (settings, TRUE);
    gtk_tree_view_set_model (GTK_TREE_VIEW (servicestree), GTK_TREE_MODEL (model_sort));

//    gtk_tree_view_set_tooltip_column (GTK_TREE_VIEW (servicestree), SERVICES_COL_TOOLTIP);
    gtk_tree_view_set_show_expanders (GTK_TREE_VIEW (servicestree), FALSE);
    gtk_tree_view_set_enable_search (GTK_TREE_VIEW (servicestree), FALSE);
    g_object_unref (G_OBJECT (model));

    column = gtk_tree_view_column_new ();

    cell_renderer = gtk_cell_renderer_pixbuf_new ();
    gtk_tree_view_column_pack_start (column, cell_renderer, FALSE);
//    gtk_tree_view_column_set_attributes (column, cell_renderer,
//                                         "pixbuf", SERVICES_COL_PIXBUF,
//                                         NULL);

//    cell_renderer = gtk_cell_renderer_text_new ();
//    gtk_tree_view_column_pack_start (column, cell_renderer, FALSE);
//    gtk_tree_view_column_set_title (column, _(titles[0]));

//    gtk_tree_view_column_set_sort_column_id (column, SERVICES_COL_NAME);
//    gtk_tree_view_column_set_resizable (column, TRUE);
//    gtk_tree_view_column_set_sizing(column, GTK_TREE_VIEW_COLUMN_FIXED);
//    gtk_tree_view_column_set_min_width (column, 1);
//    gtk_tree_view_column_set_reorderable(column, TRUE);
//
//    gsm_tree_view_append_and_bind_column (servicestree, column);
//    gtk_tree_view_set_expander_column (GTK_TREE_VIEW (servicestree), column);

    for (i = SERVICES_COL_NAME; i <= SERVICES_COL_DESCRIPTION; i++) {
        GtkTreeViewColumn *col;
        GtkCellRenderer *cell;

        cell = gtk_cell_renderer_text_new();
        col = gtk_tree_view_column_new();
        gtk_tree_view_column_pack_start(col, cell, TRUE);
        gtk_tree_view_column_set_title(col, _(titles[i]));
        gtk_tree_view_column_set_resizable(col, TRUE);
        gtk_tree_view_column_set_sizing(column, GTK_TREE_VIEW_COLUMN_AUTOSIZE);
        gtk_tree_view_column_set_sort_column_id(col, i);
        gtk_tree_view_column_set_reorderable(col, TRUE);
        gtk_tree_view_column_set_min_width(col, 20);
        gtk_tree_view_column_set_attributes(col, cell, "text", i, NULL);
        gsm_tree_view_append_and_bind_column (servicestree, col);
    }
    app->services_tree = servicestree;
    app->top_of_servicestree = NULL;
    app->last_vscroll_max_services = 0;
    app->last_vscroll_value_services = 0;

    gsm_tree_view_load_state (servicestree);

//    GtkIconTheme* theme = gtk_icon_theme_get_default();
//
    app->services_selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (servicestree));
    gtk_tree_selection_set_mode (app->services_selection, GTK_SELECTION_SINGLE);

    g_signal_connect (G_OBJECT (app->services_selection),
                      "changed",
                      G_CALLBACK (cb_services_row_selected), app);
    g_signal_connect (G_OBJECT (servicestree), "popup_menu",
                      G_CALLBACK (cb_services_tree_popup_menu), app);

    g_signal_connect (G_OBJECT (servicestree), "button_press_event",
                      G_CALLBACK (cb_services_tree_button_pressed), app);

    g_signal_connect (G_OBJECT (servicestree), "destroy",
                      G_CALLBACK (cb_servicestree_destroying),
                      app);

    g_signal_connect (G_OBJECT (servicestree), "columns-changed",
                      G_CALLBACK (cb_save_servicestree_state), app);

    g_signal_connect (G_OBJECT (model_sort), "sort-column-changed",
                      G_CALLBACK (cb_save_servicestree_state), app);

    app->settings->signal_changed(GSM_SETTING_SHOW_WHOSE_SERVICES).connect([app](const Glib::ustring& key) {
        cb_show_whose_services_changed(*app->settings.operator->(), key, app);
    });

    gtk_widget_show (GTK_WIDGET (servicestree));

    return servicestree;
}

static void
get_services_name (ServicesInfo *info,
                  const gchar *cmd, const GStrv args)
{
    if (args) {
        // look for /usr/bin/very_long_name
        // and also /usr/bin/interpreter /usr/.../very_long_name
        // which may have use prctl to alter 'cmd' name
        for (int i = 0; i != 2 && args[i]; ++i) {
            char* basename;
            basename = g_path_get_basename(args[i]);

            if (g_str_has_prefix(basename, cmd)) {
                info->name = make_string(basename);
                return;
            }

            g_free(basename);
        }
    }
    info->name = cmd;
}
static void
services_update_info_mutable_cols(ServicesInfo *info)
{
    GtkTreeModel *model;
    model = gtk_tree_model_filter_get_model (GTK_TREE_MODEL_FILTER (
            gtk_tree_model_sort_get_model(GTK_TREE_MODEL_SORT(
            gtk_tree_view_get_model (GTK_TREE_VIEW(GsmApplication::get()->services_tree))))));

    using procman::tree_store_update;

    tree_store_update(model, &info->node, SERVICES_COL_LOAD, _(info->load.c_str()));
    tree_store_update(model, &info->node, SERVICES_COL_ACTIVATE, _(info->activate.c_str()));
    tree_store_update(model, &info->node, SERVICES_COL_SUB, _(info->sub.c_str()));
//    tree_store_update(model, &info->node, SERVICES_COL_DESCRIPTION, _(info->description.c_str()));
}

static void
insert_info_to_services_tree (ServicesInfo *info, GsmApplication *app, bool forced = false)
{
    GtkTreeModel *model;
    GtkTreeModel *filtered;
    GtkTreeModel *sorted;
    sorted = gtk_tree_view_get_model (GTK_TREE_VIEW(app->services_tree));
    filtered = gtk_tree_model_sort_get_model(GTK_TREE_MODEL_SORT(sorted));
    model = gtk_tree_model_filter_get_model (GTK_TREE_MODEL_FILTER (filtered));

    gtk_tree_store_insert (GTK_TREE_STORE (model), &info->node, NULL, 0);

    gtk_tree_store_set (GTK_TREE_STORE (model), &info->node,
                        SERVICES_COL_POINTER, info,
                        SERVICES_COL_NAME, info->name.c_str(),
                        SERVICES_COL_LOAD, _(info->load.c_str()),
                        SERVICES_COL_ACTIVATE, _(info->activate.c_str()),
                        SERVICES_COL_SUB, _(info->sub.c_str()),
                        SERVICES_COL_DESCRIPTION, _(info->description.c_str()),
                        -1);

    servicesman_debug("inserted %s%s", info->name.c_str(), (forced ? " (forced)" : ""));
}

/* Removing a node with children - make sure the children are queued
** to be readded.
*/
template<typename List>
static void
remove_info_from_tree (GsmApplication *app, GtkTreeModel *model,
                       ServicesInfo& current, List &orphans, unsigned lvl = 0)
{
    GtkTreeIter child_node;

    if (std::find(orphans.begin(), orphans.end(), &current) != orphans.end()) {
        servicesman_debug("[%s] %d already removed from tree", lvl, current.name);
        return;
    }

    servicesman_debug("[%s] pid %d, %d children", lvl, current.name,
                  gtk_tree_model_iter_n_children(model, &current.node));

    // it is not possible to iterate&erase over a treeview so instead we
    // just pop one child after another and recursively remove it and
    // its children

    while (gtk_tree_model_iter_children(model, &child_node, &current.node)) {
        ServicesInfo *child = 0;
        gtk_tree_model_get(model, &child_node, SERVICES_COL_POINTER, &child, -1);
        remove_info_from_tree(app, model, *child, orphans, lvl + 1);
    }

    g_assert(not gtk_tree_model_iter_has_child(model, &current.node));

    orphans.push_back(&current);
    gtk_tree_store_remove(GTK_TREE_STORE(model), &current.node);
    procman::poison(current.node, 0x69);
}


static std::string
get_proc_kernel_wchan(glibtop_proc_kernel& obj) {
    char buf[40] = {0};
    g_strlcpy(buf, obj.wchan, sizeof(buf));
    buf[sizeof(buf)-1] = '\0';
    return buf;
}

ServicesInfo::ServicesInfo(std::string name)
{
    this->name = name;
}

static void
refresh_services_list (GsmApplication *app, ServicesList * services_list, const guint n)
{
    typedef std::list<ServicesInfo*> ServicesList;
    ServicesList add_list;

    GtkTreeModel    *model = gtk_tree_model_filter_get_model (GTK_TREE_MODEL_FILTER (
                             gtk_tree_model_sort_get_model(GTK_TREE_MODEL_SORT (
                             gtk_tree_view_get_model (GTK_TREE_VIEW(app->services_tree))))));

    // Add or update services in the services list
    auto iter = services_list->begin();
    for(; iter != services_list->end(); ++iter) {
        ServicesInfo *info = & iter->second;
        ServicesInfo *find_info = app->services.find(info->name);
        if ( find_info == NULL) {
            info = app->services.add(info);
            add_list.push_back(info);
//            g_print("Insert!%s\n", info->name);
        } else {
            * find_info = iter->second;
        }
    }

    for (auto& v : add_list) insert_info_to_services_tree(v, app);

    for (auto& v : app->services) services_update_info_mutable_cols(&v.second);
}

static
std::list<string> split(const string &str,const string &pattern)
{
    //const char* convert to char*
    char * strc = new char[strlen(str.c_str())+1];
    strcpy(strc, str.c_str());
    std::list<string> resultVec;
    char* tmpStr = strtok(strc, pattern.c_str());
    while (tmpStr != NULL)
    {
        resultVec.push_back(string(tmpStr));
        tmpStr = strtok(NULL, pattern.c_str());
    }

    delete[] strc;

    return resultVec;
}

void
servicestable_update (GsmApplication *app)
{
    ServicesList * services_list = new ServicesList();
    sd_bus_message * msg = NULL;

    auto whose_services = app->settings->get_string(GSM_SETTING_SHOW_WHOSE_SERVICES);

    int size = get_list_units(&services_list, &msg, whose_services);
    if ( size <= 0) {
        return ;
    }

    refresh_services_list (app, services_list, size);

    if (services_list) {
        delete services_list;
    }
//    if ( msg) {
//        delete msg;
//    }
    // juggling with tree scroll position to fix https://bugzilla.gnome.org/show_bug.cgi?id=92724
    GtkTreePath* current_top;
    if (gtk_tree_view_get_path_at_pos(GTK_TREE_VIEW(app->services_tree), 0,0, &current_top, NULL, NULL, NULL)) {
        GtkAdjustment *vadjustment = GTK_ADJUSTMENT (gtk_scrollable_get_vadjustment (GTK_SCROLLABLE (app->services_tree)));
        gdouble current_max = gtk_adjustment_get_upper(vadjustment);
        gdouble current_value = gtk_adjustment_get_value(vadjustment);

        if (app->top_of_servicestree) {
            // if the visible cell from the top of the tree is still the same, as last time 
            if (gtk_tree_path_compare (app->top_of_servicestree, current_top) == 0) {
                //but something from the scroll parameters has changed compared to the last values
                if (app->last_vscroll_value_services == 0 && current_value != 0) {
                    // the tree was scrolled to top, and something has been added above the current top row
                    gtk_tree_view_scroll_to_point(GTK_TREE_VIEW(app->services_tree), -1, 0);
                } else if (current_max > app->last_vscroll_max_services && app->last_vscroll_max_services == app->last_vscroll_value_services) {
                    // the tree was scrolled to bottom, something has been added below the current bottom row
                    gtk_tree_view_scroll_to_point(GTK_TREE_VIEW(app->services_tree), -1, current_max);
                }
            }

            gtk_tree_path_free(app->top_of_servicestree);
        }

        app->top_of_servicestree = current_top;
        app->last_vscroll_value_services = current_value;
        app->last_vscroll_max_services = current_max;
    }
}

void
servicestable_free_table (GsmApplication * const app)
{
    app->services.clear();
}

void
servicestable_freeze (GsmApplication *app)
{
    if (app->timeout) {
      g_source_remove (app->timeout);
      app->timeout = 0;
    }
}

void
servicestable_thaw (GsmApplication *app)
{
    if (app->timeout)
        return;

    app->timeout = g_timeout_add (app->services_config.update_interval,
                                  cb_services_timeout,
                                  app);
}