//
// Created by cuilei on 2020/3/26.
//

#include <servicesutil.h>
#include <systemd-interfaces.h>
#include <stdarg.h>
#include <malloc.h>

static BusTransport arg_transport = BUS_TRANSPORT_LOCAL;
static const char *arg_host = NULL;
static sd_bus *buses[_BUS_FOCUS_MAX] = {};
static char **arg_states = NULL;
static UnitFileScope arg_scope = UNIT_FILE_SYSTEM;
static bool arg_ask_password = false;


bool size_multiply_overflow(size_t size, size_t need) {
    return _unlikely_(need != 0 && size > (SIZE_MAX / need));
}

void *malloc_multiply(size_t size, size_t need) {
    if (size_multiply_overflow(size, need))
        return NULL;

    return malloc(size * need ?: 1);
}

void* greedy_realloc(void **p, size_t *allocated, size_t need, size_t size) {
    size_t a, newalloc;
    void *q;

    assert(p);
    assert(allocated);

    if (*allocated >= need)
        return *p;

    if (_unlikely_(need > SIZE_MAX/2)) /* Overflow check */
        return NULL;

    newalloc = need * 2;
    if (size_multiply_overflow(newalloc, size))
        return NULL;

    a = newalloc * size;
    if (a < 64) /* Allocate at least 64 bytes */
        a = 64;

    q = realloc(*p, a);
    if (!q)
        return NULL;

    if (size > 0) {
        size_t bn;

        /* Adjust for the 64 byte minimum */
        newalloc = a / size;

        bn = malloc_usable_size(q) / size;
        if (bn > newalloc) {
            void *qq;

            /* The actual size allocated is larger than what we asked for. Let's call realloc() again to
             * take possession of the extra space. This should be cheap, since libc doesn't have to move
             * the memory for this. */

            qq = realloc(q, bn * size);
            if (_likely_(qq)) {
                *p = qq;
                *allocated = bn;
                return qq;
            }
        }
    }

    *p = q;
    *allocated = newalloc;
    return q;
}



char *strjoin_real(const char *x, ...) {
    va_list ap;
    size_t l;
    char *r, *p;

    va_start(ap, x);

    if (x) {
        l = strlen(x);

        for (;;) {
            const char *t;
            size_t n;

            t = va_arg(ap, const char *);
            if (!t)
                break;

            n = strlen(t);
            if (n > ((size_t) -1) - l) {
                va_end(ap);
                return NULL;
            }

            l += n;
        }
    } else
        l = 0;

    va_end(ap);

    r = service_new(char, l+1);
    if (!r)
        return NULL;

    if (x) {
        p = stpcpy(r, x);

        va_start(ap, x);

        for (;;) {
            const char *t;

            t = va_arg(ap, const char *);
            if (!t)
                break;

            p = stpcpy(p, t);
        }

        va_end(ap);
    } else
        r[0] = 0;

    return r;
}



static int parse_boolean(const char *v) {
    if (!v)
        return -EINVAL;

    if (streq(v, "1") || strcaseeq(v, "yes") || strcaseeq(v, "y") || strcaseeq(v, "true") || strcaseeq(v, "t") || strcaseeq(v, "on"))
        return 1;
    else if (streq(v, "0") || strcaseeq(v, "no") || strcaseeq(v, "n") || strcaseeq(v, "false") || strcaseeq(v, "f") || strcaseeq(v, "off"))
        return 0;

    return -EINVAL;
}

static int getenv_bool(const char *p) {
    const char *e;

    e = getenv(p);
    if (!e)
        return -ENXIO;

    return parse_boolean(e);
}

static bool pid_is_valid(pid_t p) {
    return p > 0;
}

int getpeercred(int fd, struct ucred *ucred) {
    socklen_t n = sizeof(struct ucred);
    struct ucred u;
    int r;

    assert(fd >= 0);
    assert(ucred);

    r = getsockopt(fd, SOL_SOCKET, SO_PEERCRED, &u, &n);
    if (r < 0)
        return -errno;

    if (n != sizeof(struct ucred))
        return -EIO;

    /* Check if the data is actually useful and not suppressed due to namespacing issues */
    if (!pid_is_valid(u.pid))
        return -ENODATA;

    /* Note that we don't check UID/GID here, as namespace translation works differently there: instead of
     * receiving in "invalid" user/group we get the overflow UID/GID. */

    *ucred = u;
    return 0;
}

static int bus_check_peercred(sd_bus *c) {
    struct ucred ucred;
    int fd, r;

    assert(c);

    fd = sd_bus_get_fd(c);
    if (fd < 0)
        return fd;

    r = getpeercred(fd, &ucred);
    if (r < 0)
        return r;

    if (ucred.uid != 0 && ucred.uid != geteuid())
        return -EPERM;

    return 1;
}

static int bus_connect_system_systemd(sd_bus **_bus) {
    _cleanup_(sd_bus_close_unrefp) sd_bus *bus = NULL;
    int r;

    assert(_bus);

    if (geteuid() != 0)
        return sd_bus_default_system(_bus);

    /* If we are root then let's talk directly to the system
     * instance, instead of going via the bus */

    r = sd_bus_new(&bus);
    if (r < 0)
        return r;

    r = sd_bus_set_address(bus, "unix:path=/run/systemd/private");
    if (r < 0)
        return r;

    r = sd_bus_start(bus);
    if (r < 0)
        return sd_bus_default_system(_bus);

    r = bus_check_peercred(bus);
    if (r < 0)
        return r;

    *_bus = bus;

    return 0;
}

int sd_booted(void) {
    /* We test whether the runtime unit file directory has been
     * created. This takes place in mount-setup.c, so is
     * guaranteed to happen very early during boot. */

    if (laccess("/run/systemd/system/", F_OK) >= 0)
        return true;

    if (errno == ENOENT)
        return false;

    return -errno;
}

char hexchar(int x) {
    static const char table[17] = "0123456789abcdef";

    return table[x & 15];
}

char *bus_address_escape(const char *v) {
    const char *a;
    char *r, *b;

    r = service_new(char, strlen(v)*3+1);
    if (!r)
        return NULL;

    for (a = v, b = r; *a; a++) {

        if ((*a >= '0' && *a <= '9') ||
            (*a >= 'a' && *a <= 'z') ||
            (*a >= 'A' && *a <= 'Z') ||
            strchr("_-/.", *a))
            *(b++) = *a;
        else {
            *(b++) = '%';
            *(b++) = hexchar(*a >> 4);
            *(b++) = hexchar(*a & 0xF);
        }
    }

    *b = 0;
    return r;
}

int bus_connect_user_systemd(sd_bus **_bus) {
    _cleanup_(sd_bus_close_unrefp) sd_bus *bus = NULL;
    _cleanup_free_ char *ee = NULL;
    const char *e;
    int r;

    assert(_bus);

    e = secure_getenv("XDG_RUNTIME_DIR");
    if (!e)
        return sd_bus_default_user(_bus);

    ee = bus_address_escape(e);
    if (!ee)
        return -ENOMEM;

    r = sd_bus_new(&bus);
    if (r < 0)
        return r;

    bus->address = strjoin("unix:path=", ee, "/systemd/private");
    if (!bus->address)
        return -ENOMEM;

    r = sd_bus_start(bus);
    if (r < 0)
        return sd_bus_default_user(_bus);

    r = bus_check_peercred(bus);
    if (r < 0)
        return r;

    *_bus = bus;

    return 0;
}

static int bus_connect_transport_systemd(BusTransport transport, const char *host, bool user, sd_bus **bus) {
    int r;

    assert(transport >= 0);
    assert(transport < _BUS_TRANSPORT_MAX);
    assert(bus);

    switch (transport) {

        case BUS_TRANSPORT_LOCAL:
            if (user)
                r = bus_connect_user_systemd(bus);
            else {
                if (sd_booted() <= 0) {
                    /* g_print a friendly message when the local system is actually not running systemd as PID 1. */
                    return -1;
                }
                r = bus_connect_system_systemd(bus);
            }
            break;

        case BUS_TRANSPORT_REMOTE:
            r = sd_bus_open_system_remote(bus, host);
            break;

        case BUS_TRANSPORT_MACHINE:
            r = sd_bus_open_system_machine(bus, host);
            break;

        default:
            r = -1;
    }

    return r;
}

int bus_connect_transport(BusTransport transport, const char *host, bool user, sd_bus **ret) {
    _cleanup_(sd_bus_close_unrefp) sd_bus *bus = NULL;
    int r;

    assert(transport >= 0);
    assert(transport < _BUS_TRANSPORT_MAX);
    assert(ret);

    switch (transport) {

        case BUS_TRANSPORT_LOCAL:
            if (user)
                r = sd_bus_default_user(&bus);
            else {
                if (sd_booted() <= 0) {
                    /* g_print a friendly message when the local system is actually not running systemd as PID 1. */
                    g_print("System has not been booted with systemd as init system (PID 1). Can't operate.");

                    return -EHOSTDOWN;
                }
                r = sd_bus_default_system(&bus);
            }
            break;

        case BUS_TRANSPORT_REMOTE:
            r = sd_bus_open_system_remote(&bus, host);
            break;

        case BUS_TRANSPORT_MACHINE:
            r = sd_bus_open_system_machine(&bus, host);
            break;

        default:
            r = -1;
    }
    if (r < 0)
        return r;

    r = sd_bus_set_exit_on_disconnect(bus, true);
    if (r < 0)
        return r;

    *ret = bus;

    return 0;
}

static int acquire_bus(BusFocus focus, sd_bus **ret) {
    int r;

    assert(focus < _BUS_FOCUS_MAX);
    assert(ret);

    /* We only go directly to the manager, if we are using a local transport */
    if (arg_transport != BUS_TRANSPORT_LOCAL)
        focus = BUS_FULL;

    if (getenv_bool("SYSTEMCTL_FORCE_BUS") > 0)
        focus = BUS_FULL;

    if (!buses[focus]) {
        bool user;

        user = arg_scope != UNIT_FILE_SYSTEM;

        if (focus == BUS_MANAGER)
            r = bus_connect_transport_systemd(arg_transport, arg_host, user, &buses[focus]);
        else
            r = bus_connect_transport(arg_transport, arg_host, user, &buses[focus]);
        if (r < 0)
            g_print("Failed to connect to bus: %d", r);

        (void) sd_bus_set_allow_interactive_authorization(buses[focus], arg_ask_password);
    }

    *ret = buses[focus];
    return 0;
}

int bus_parse_unit_info(sd_bus_message *message, ServicesInfo **u) {
    assert(message);
    assert(u);
    return sd_bus_message_read(
            message,
            "(ssssssouso)",
            &(*u)->name,
            &(*u)->description,
            &(*u)->load,
            &(*u)->activate,
            &(*u)->sub,
            &(*u)->following,
            &(*u)->unit_path,
            &(*u)->job_id,
            &(*u)->job_type,
            &(*u)->job_path);
}

bool strv_isempty(char * const *l) {
    return !l || !*l;
}

char **strv_skip(char **l, size_t n) {

    while (n > 0) {
        if (strv_isempty(l))
            return l;

        l++, n--;
    }

    return l;
}

int get_list_units(
        ServicesList ** unit_infos,
        sd_bus_message **_reply,
        std::string tag) {

    sd_bus * bus;
    int c = 0;
    size_t size = c;
    int r;

    _cleanup_(sd_bus_message_unrefp) sd_bus_message *m = NULL;
    _cleanup_(sd_bus_error_free) sd_bus_error error = SD_BUS_ERROR_NULL;
    _cleanup_(sd_bus_message_unrefp) sd_bus_message *reply = NULL;

    r = acquire_bus(BUS_MANAGER, &bus);
    if (r < 0)
        return r;
    assert(bus);
    assert(unit_infos);
    assert(_reply);

    r = sd_bus_message_new_method_call(
            bus,
            &m,
            "org.freedesktop.systemd1",
            "/org/freedesktop/systemd1",
            "org.freedesktop.systemd1.Manager",
            "ListUnits");
    if (r < 0)
        return r;

    r = sd_bus_call(bus, m, 0, &error, &reply);
    if (r < 0 && (sd_bus_error_has_name(&error, SD_BUS_ERROR_UNKNOWN_METHOD) ||
                  sd_bus_error_has_name(&error, SD_BUS_ERROR_ACCESS_DENIED))) {
        /* Fallback to legacy ListUnitsFiltered method */
        g_print("Failed to list units: %s Falling back to ListUnitsFiltered method.", error.message);
        m = sd_bus_message_unref(m);
        sd_bus_error_free(&error);

        r = sd_bus_message_new_method_call(
                bus,
                &m,
                "org.freedesktop.systemd1",
                "/org/freedesktop/systemd1",
                "org.freedesktop.systemd1.Manager",
                "ListUnitsFiltered");
        if (r < 0)
            return r;

        r = sd_bus_message_append_strv(m, arg_states);
        if (r < 0)
            return r;

        r = sd_bus_call(bus, m, 0, &error, &reply);
    }
    if (r < 0)
        return r;

    r = sd_bus_message_enter_container(reply, SD_BUS_TYPE_ARRAY, "(ssssssouso)");
    if (r < 0)
        return r;

    *_reply = reply;

    ServicesInfo * u = new ServicesInfo();
    while (sd_bus_message_read(
            reply,
            "(ssssssouso)",
            &u->name_service_str,
            &u->description_str,
            &u->load_str,
            &u->activate_str,
            &u->sub_str,
            &u->following_str,
            &u->unit_path_str,
            &u->job_id,
            &u->job_type_str,
            &u->job_path_str) > 0) {

//        if (!GREEDY_REALLOC(*unit_infos, size, c+1))
//            return -1;

        u->name_service = std::string(u->name_service_str);
        u->name = std::string(u->name_service_str);
        u->description = std::string(u->description_str);
        u->load = std::string(u->load_str);
        u->activate = std::string(u->activate_str);
        u->sub = std::string(u->sub_str);
        u->following = std::string(u->following_str);
        u->unit_path = std::string(u->unit_path_str);
        u->job_type = std::string(u->job_type_str);
        u->job_path = std::string(u->job_path_str);

        if ( strstr(u->name.c_str(), ".service") == NULL) {
            continue;
        } else {
            u->name = u->name_service.substr(0, u->name_service.find_first_of('.', 0));
//            u->name = u->name_service;
            if ( tag != "all") {
                if ( strcmp ( u->activate.c_str(), tag.c_str()) == 0 ||
                     strcmp ( u->sub.c_str(), tag.c_str()) == 0) {
                    (*unit_infos)->add(u);
                    c++;
//                    g_print("[all]id:%s\n", u->name_service.c_str());
                }
            } else {
//                g_print("id:%s\n", u->name.c_str());
                (*unit_infos)->add(u);
                c++;
            }
        }
    }
    delete u;
    if (r < 0)
        return r;

    r = sd_bus_message_exit_container(reply);
    if (r < 0)
        return r;

    return c;
}

#define _g_object_unref0(var) ((var == NULL) ? NULL : (var = (g_object_unref (var), NULL)))
#define _g_error_free0(var) ((var == NULL) ? NULL : (var = (g_error_free (var), NULL)))

static
void start_service_util (std::string unit_path, GError** error)
{
    Unit *unit = NULL;
    GBusType type = 1;
    unit = (Unit *) g_initable_new(TYPE_UNIT_PROXY, NULL, error, "g-flags", 0, "g-name", "org.freedesktop.systemd1",
                                   "g-bus-type", type, "g-object-path", (const gchar *) unit_path.c_str(), "g-interface-name",
                                   "org.freedesktop.systemd1.Unit", "g-interface-info",
                                   g_type_get_qdata(TYPE_UNIT, g_quark_from_static_string("vala-dbus-interface-info")),
                                   NULL);
    unit_start (unit, "replace", error);
    _g_error_free0(*error);
    _g_object_unref0(unit) ;
}

static
void stop_service_util(std::string unit_path,
              GError** error) {
    Unit *unit = NULL;
    GBusType type = 1;
    unit = (Unit *) g_initable_new(TYPE_UNIT_PROXY, NULL, error, "g-flags", 0, "g-name", "org.freedesktop.systemd1",
                                   "g-bus-type", type, "g-object-path", (const gchar *) unit_path.c_str(), "g-interface-name",
                                   "org.freedesktop.systemd1.Unit", "g-interface-info",
                                   g_type_get_qdata(TYPE_UNIT, g_quark_from_static_string("vala-dbus-interface-info")),
                                   NULL);
    unit_stop (unit, "replace", error);
    _g_error_free0(*error);
    _g_object_unref0(unit) ;
}

static
void restart_service_util (std::string unit_path,
                   GError** error)
{
    Unit *unit = NULL;
    GBusType type = 1;
    unit = (Unit *) g_initable_new(TYPE_UNIT_PROXY, NULL, error, "g-flags", 0, "g-name", "org.freedesktop.systemd1",
                                   "g-bus-type", type, "g-object-path", (const gchar *) unit_path.c_str(), "g-interface-name",
                                   "org.freedesktop.systemd1.Unit", "g-interface-info",
                                   g_type_get_qdata(TYPE_UNIT, g_quark_from_static_string("vala-dbus-interface-info")),
                                   NULL);
    unit_restart (unit, "replace", error);
    _g_error_free0(*error);
    _g_object_unref0(unit) ;
}

//static
//int enable_unit(std::string unit_path,
//                GError** error) {
//    Unit *unit = NULL;
//    GBusType type = 0;
//    unit = (Unit *) g_initable_new(TYPE_UNIT_PROXY, NULL, error, "g-flags", 0, "g-name", "org.freedesktop.systemd1",
//                                   "g-bus-type", type, "g-object-path", (const gchar *) unit_path.c_str(), "g-interface-name",
//                                   "org.freedesktop.systemd1.Unit", "g-interface-info",
//                                   g_type_get_qdata(TYPE_UNIT, g_quark_from_static_string("vala-dbus-interface-info")),
//                                   NULL);
//    unit_enable (unit, "replace", error);
//}
//
//static
//int disable_unit(std::string unit_path,
//                 GError** error) {
//    Unit *unit = NULL;
//    GBusType type = 0;
//    unit = (Unit *) g_initable_new(TYPE_UNIT_PROXY, NULL, error, "g-flags", 0, "g-name", "org.freedesktop.systemd1",
//                                   "g-bus-type", type, "g-object-path", (const gchar *) unit_path.c_str(), "g-interface-name",
//                                   "org.freedesktop.systemd1.Unit", "g-interface-info",
//                                   g_type_get_qdata(TYPE_UNIT, g_quark_from_static_string("vala-dbus-interface-info")),
//                                   NULL);
//    unit_start (unit, "replace", error);
//}